(* Persistent queue *)

open Lazy;;

module Stream =
  struct
    type 'a cell = Nil | Cons of 'a * 'a stream
    and 'a stream = 'a cell lazy_t

    let empty = lazy Nil;;

    let iter f u =
      let rec loop v =
        match force v with
        | Cons(x,w) -> f x; loop w
        | Nil -> ()
      in
      loop u
    ;;

    let rec add x u = lazy (Cons(x, u));;

    let rec append u v =
      match force u with
      | Nil -> v
      | Cons(x,w) -> lazy (Cons(x, append w v))
    ;;
    let rec prefix i u =
      if i = 0 then
        empty
      else
        match force u with
        | Nil -> empty
        | Cons(x,v) -> lazy (Cons(x, prefix (i - 1) v))
    ;;

    let rec remove_prefix (i : int) (u : 'a stream) =
      if i = 0 then
        u
      else
        match force u with
        | Nil -> empty
        | Cons(_, v) -> remove_prefix (i - 1) v
    ;;

    let reverse u =
      let rec loop u r =
        match force u with
        | Nil -> r
        | Cons(x, v) -> loop v (lazy (Cons(x, r)))
      in
      loop u empty
    ;;
  end
;;

open Stream;;

type 'a t = 'a stream * 'a list * 'a stream

exception Empty;;

let empty = (empty, [], empty);;
let is_empty (f, _, _) =
  match force f with
  | Nil -> true
  | _ -> false
;;
let rec rotate (f,r,s) =
  match (force f, r, s) with
  | (Nil, y :: _, a) -> lazy (Cons(y, a))
  | (Cons(x, xs), y :: ys, a) -> lazy (Cons(x, rotate (xs, ys, lazy (Cons(y,a)))))
  | _ -> assert false (* XXX *)
;;
let exec (f, r, s) =
  match force s with
  | Cons(x,t) -> (f, r, t)
  | Nil ->
      let f' = rotate (f, r, Stream.empty) in
      (f', [], f')
;;
let add x (f,r,s) =
  exec (f, x :: r, s);;

let head (f, r, s) =
  match force f with
  | Nil -> raise Empty
  | Cons(x,_) -> x
;;

let tail (f, r, s) =
  match force f with
  | Nil -> raise Empty
  | Cons(_,f') -> exec (f', r, s)
;;

let take q = (head q, tail q);;
